import base64
import datetime
import pytz
import re
from contextlib import contextmanager
from dataclasses import dataclass
from io import StringIO
from time import sleep
from typing import Optional, Tuple, Union, Any, List

import agate
import dbt.clients.agate_helper
from clickzetta.dbapi.connection import Connection as ClickZettaConnection
from clickzetta.client import Client

from dbt.exceptions import (
    DbtInternalError,
    DbtRuntimeError,
    FailedToConnectError,
    DbtDatabaseError,
    DbtProfileError,
)
from dbt.adapters.base import Credentials  # type: ignore
from dbt.contracts.connection import AdapterResponse, Connection
from dbt.adapters.sql import SQLConnectionManager  # type: ignore
from dbt.events import AdapterLogger  # type: ignore
from dbt.events.functions import warn_or_error
from dbt.events.types import AdapterEventWarning
from dbt.ui import line_wrap_message, warning_tag

logger = AdapterLogger("ClickZetta")
ROW_VALUE_REGEX = re.compile(r"Row Values: \[.*\]")


@dataclass
class ClickZettaAdapterResponse(AdapterResponse):
    query_id: str = ""


@dataclass
class ClickZettaCredentials(Credentials):
    database: Optional[str] = ""  # type: ignore
    workspace: Optional[str] = ""
    instance: Optional[str] = ""
    vcluster: Optional[str] = "default"
    password: Optional[str] = ""
    service: Optional[str] = ""
    username: Optional[str] = ""
    schema: Optional[str] = "public"
    connect_retries: int = 3
    reuse_connections: bool = True
    split_size: Optional[int] = 64 * 1024 * 1024

    @classmethod
    def __pre_deserialize__(cls, data):
        data = super().__pre_deserialize__(data)
        if "workspace" not in data:
            raise DbtProfileError(
                "The 'workspace' field is required for ClickZetta connections."
            )
        if "database" not in data:
            data["database"] = data["workspace"]
        return data

    @property
    def type(self):
        return "clickzetta"

    @property
    def unique_field(self):
        return self.username

    def _connection_keys(self):
        return (
            "workspace",
            "instance",
            "vcluster",
            "username",
            "schema",
            "password",
            "service",
            "database",
        )


class ClickZettaConnectionManager(SQLConnectionManager):
    TYPE = "clickzetta"

    @contextmanager
    def exception_handler(self, sql):
        try:
            yield
        except Exception as e:
            logger.warning("Error running SQL: {}", sql)
            logger.warning("Exception {}", e)
            logger.debug("Rolling back transaction.")
            self.rollback_if_open()
            if isinstance(e, DbtRuntimeError):
                # during a sql query, an internal to dbt exception was raised.
                # this sounds a lot like a signal handler and probably has
                # useful information, so raise it without modification.
                raise
            raise DbtRuntimeError(str(e)) from e

    @classmethod
    def open(cls, connection):
        if connection.state == "open":
            logger.debug("Connection is already open, skipping open.")
            return connection

        creds = connection.credentials

        def connect():
            session_parameters = {}
            client = Client(
                username=creds.username,
                password=creds.password,
                instance=creds.instance,
                workspace=creds.workspace,
                vcluster=creds.vcluster,
                schema=creds.schema,
                service=creds.service,
            )
            handle = ClickZettaConnection(client=client)

            return handle

        def exponential_backoff(attempt: int):
            return attempt * attempt

        return cls.retry_connection(
            connection,
            connect=connect,
            logger=logger,
            retry_limit=creds.connect_retries,
            retry_timeout=exponential_backoff,
            retryable_exceptions=[],
        )

    def cancel(self, connection):
        pass

    @classmethod
    def get_response(cls, cursor) -> ClickZettaAdapterResponse:
        code = cursor.rowcount
        logger.debug(f"code: {code}")
        if code is not None:
            code = "SUCCESS"

        return ClickZettaAdapterResponse(
            _message="{} {}".format(code, cursor.rowcount),
            rows_affected=cursor.rowcount,
            code=code,
        )  # type: ignore

    def add_begin_query(self, *args, **kwargs):
        pass

    def add_commit_query(self, *args, **kwargs):
        pass

    def begin(self):
        pass

    def commit(self):
        pass

    def clear_transaction(self):
        pass

    @classmethod
    def _split_queries(cls, sql):
        pass

    def execute(
            self, sql: str, auto_begin: bool = False, fetch: bool = False, limit: Optional[int] = None
    ) -> Tuple[AdapterResponse, agate.Table]:
        _, cursor = self.add_query(sql, auto_begin)
        logger.debug(f"dbt_execute_sql: {sql}")
        response = self.get_response(cursor)
        if fetch:
            table = self.get_result_from_cursor(cursor)
        else:
            table = dbt.clients.agate_helper.empty_table()
        return response, table

    def add_standard_query(self, sql: str, **kwargs) -> Tuple[Connection, Any]:
        return super().add_query(self._add_query_comment(sql), **kwargs)


    def add_query(
            self,
            sql: str,
            auto_begin: bool = True,
            bindings: Optional[Any] = None,
            abridge_sql_log: bool = False,
    ) -> Tuple[Connection, Any]:  # type: ignore
        # TODO(hanmiao.li): clickzetta cursor is not support pass bingdings, will support later.
        if bindings:
            cast_bindings = []
            for binding in bindings:
                cast_bindings.append(f"'{str(binding)}'")
            sql = sql % tuple(cast_bindings)
            bindings = None

        bindings = {'hints': {'cz.mapper.file.split.size': self.get_thread_connection().credentials.split_size}}

        connection, cursor = self._add_standard_queries(
            [sql],
            auto_begin=auto_begin,
            bindings=bindings,
            abridge_sql_log=abridge_sql_log,
        )

        if cursor is None:
            self._raise_cursor_not_found_error(sql)

        return connection, cursor  # type: ignore

    def _add_standard_queries(self, queries: List[str], **kwargs) -> Tuple[Connection, Any]:
        for query in queries:
            connection, cursor = self.add_standard_query(query, **kwargs)
        return connection, cursor

    def _raise_cursor_not_found_error(self, sql: str):
        conn = self.get_thread_connection()
        try:
            conn_name = conn.name
        except AttributeError:
            conn_name = None

        raise DbtRuntimeError(
            f"""Tried to run an empty query on model '{conn_name or "<None>"}'. If you are """
            f"""conditionally running\nsql, e.g. in a model hook, make """
            f"""sure your `else` clause contains valid sql!\n\n"""
            f"""Provided SQL:\n{sql}"""
        )

    def release(self):
        if self.profile.credentials.reuse_connections:  # type: ignore
            return
        super().release()



from dataclasses import dataclass, field
from dbt.adapters.base.relation import BaseRelation, Policy
from dbt.exceptions import DbtRuntimeError


@dataclass
class ClickZettaQuotePolicy(Policy):
    database: bool = False
    schema: bool = False
    identifier: bool = False

@dataclass
class ClickZettaIncludePolicy(Policy):
    database: bool = False
    schema: bool = True
    identifier: bool = True


@dataclass(frozen=True, eq=False, repr=False)
class ClickZettaRelation(BaseRelation):
    quote_policy: Policy = field(default_factory=lambda: ClickZettaQuotePolicy())
    include_policy: Policy = field(default_factory=lambda: ClickZettaIncludePolicy())
    quote_character: str = "`"

    # def __post_init__(self):
    #     if self.database != self.schema and self.database:
    #         raise DbtRuntimeError("Cannot set database in clickzetta!")

    def render(self):
        return super().render()



from dataclasses import dataclass
from typing import Any, Dict, Optional, TypeVar, Union

from dbt.adapters.base.column import Column
from dbt.exceptions import DbtRuntimeError
from hologram import JsonDict
from dbt.dataclass_schema import dbtClassMixin


@dataclass
class ClickZettaColumn(dbtClassMixin, Column):  # type: ignore
    table_database: Optional[str] = None
    table_schema: Optional[str] = None
    table_name: Optional[str] = None

    @classmethod
    def translate_type(cls, dtype: str) -> str:
        return dtype

    def is_integer(self) -> bool:
        return self.dtype.lower() in [
            "int8",
            "int16",
            "int32",
            "int64",
        ]

    def is_float(self):
        return self.dtype.lower() in [
            "float32",
            "float64",
            # TODO(hanmiao.li): decimal is a subclass of float, but we don't want to treat it
            # "decimal",
        ]

    @property
    def quoted(self) -> str:
        return "`{}`".format(self.column)

    @property
    def data_type(self) -> str:
        return self.dtype

    def is_string(self) -> bool:
        return self.dtype.lower() in [
            "string",
            "varchar",
            "char",
        ]

    def to_column_dict(self, omit_none: bool = True, validate: bool = False) -> JsonDict:
        original_dict = self.to_dict(omit_none=omit_none)
        return original_dict



from dataclasses import dataclass
from typing import Mapping, Any, Optional, List, Union, Dict, Set
from concurrent.futures import Future

import agate
import dbt.exceptions

from dbt.adapters.base.impl import AdapterConfig, ConstraintSupport  # type: ignore
from dbt.adapters.sql import SQLAdapter  # type: ignore
from dbt.adapters.sql.impl import (
    LIST_SCHEMAS_MACRO_NAME,
    LIST_RELATIONS_MACRO_NAME,
)
from dbt.adapters.base.impl import catch_as_completed, ConstraintSupport

from dbt.adapters.clickzetta import ClickZettaConnectionManager
from dbt.adapters.clickzetta import ClickZettaRelation
from dbt.adapters.clickzetta import ClickZettaColumn
from dbt.contracts.graph.manifest import Manifest
from dbt.contracts.graph.nodes import ConstraintType
from dbt.contracts.relation import RelationType
from dbt.exceptions import CompilationError, DbtDatabaseError, DbtRuntimeError
from dbt.adapters.base import BaseRelation, SchemaSearchMap
from dbt.clients.agate_helper import DEFAULT_TYPE_TESTER
from dbt.contracts.connection import AdapterResponse
from dbt.contracts.graph.nodes import ConstraintType
from dbt.contracts.relation import RelationType
from dbt.events import AdapterLogger
from dbt.utils import executor, AttrDict

GET_COLUMNS_IN_RELATION_RAW_MACRO_NAME = "get_columns_in_relation_raw"
LIST_SCHEMAS_MACRO_NAME = "list_schemas"
LIST_RELATIONS_MACRO_NAME = "list_relations_without_caching"
LIST_RELATIONS_SHOW_TABLES_MACRO_NAME = "list_relations_show_tables_without_caching"
DESCRIBE_TABLE_EXTENDED_MACRO_NAME = "describe_table_extended_without_caching"
DROP_RELATION_MACRO_NAME = "drop_relation"

TABLE_OR_VIEW_NOT_FOUND_MESSAGES = (
    "[TABLE_OR_VIEW_NOT_FOUND]",
    "Table or view not found",
    "NoSuchTableException",
)
logger = AdapterLogger(__name__)


@dataclass
class ClickZettaConfig(AdapterConfig):
    pass


class ClickZettaAdapter(SQLAdapter):
    Relation = ClickZettaRelation
    Column = ClickZettaColumn
    ConnectionManager = ClickZettaConnectionManager

    AdapterSpecificConfigs = ClickZettaConfig

    CONSTRAINT_SUPPORT = {
        ConstraintType.check: ConstraintSupport.NOT_SUPPORTED,
        ConstraintType.not_null: ConstraintSupport.ENFORCED,
        ConstraintType.unique: ConstraintSupport.NOT_ENFORCED,
        ConstraintType.primary_key: ConstraintSupport.NOT_ENFORCED,
        ConstraintType.foreign_key: ConstraintSupport.NOT_SUPPORTED,
    }

    @classmethod
    def date_function(cls):
        return "current_timestamp()"

    @classmethod
    def _catalog_filter_table(cls, table: agate.Table, manifest: Manifest) -> agate.Table:
        lowered = table.rename(column_names=[c.lower() for c in table.column_names])
        return super()._catalog_filter_table(lowered, manifest)

    def _get_catalog_schemas(self, manifest: Manifest) -> SchemaSearchMap:
        candidates = super()._get_catalog_schemas(manifest)
        db_schemas: Dict[str, Set[str]] = {}
        result = SchemaSearchMap()

        for candidate, schemas in candidates.items():
            database = candidate.database
            if database not in db_schemas:
                db_schemas[database] = set(self.list_schemas(database))  # type: ignore[index]
            if candidate.schema in db_schemas[database]:  # type: ignore[index]
                result[candidate] = schemas
            else:
                logger.debug(
                    "Skipping catalog for {}.{} - schema does not exist".format(
                        database, candidate.schema
                    )
                )
        return result

    @classmethod
    def convert_text_type(cls, agate_table, col_idx):
        return "string"

    @classmethod
    def convert_number_type(cls, agate_table, col_idx):
        decimals = agate_table.aggregate(agate.MaxPrecision(col_idx))
        return "double" if decimals else "bigint"

    @classmethod
    def convert_date_type(cls, agate_table, col_idx):
        return "date"

    @classmethod
    def convert_time_type(cls, agate_table, col_idx):
        return "time"

    @classmethod
    def convert_datetime_type(cls, agate_table, col_idx):
        return "timestamp"

    def quote(self, identifier):
        return "`{}`".format(identifier)

    def parse_describe_extended(
            self, relation: BaseRelation, raw_rows: AttrDict
    ) -> List[ClickZettaColumn]:
        # Convert the Row to a dict
        dict_rows = [dict(zip(row._keys, row._values)) for row in raw_rows]

        rows = [row for row in dict_rows if not row["column_name"].startswith("#")]

        return [
            ClickZettaColumn(
                table_database=None,
                table_schema=relation.schema,
                table_name=relation.name,
                column=column["column_name"],
                dtype=column["data_type"],
            )
            for idx, column in enumerate(rows)
        ]

    def get_columns_in_relation(self, relation: BaseRelation) -> List[ClickZettaColumn]:
        columns = []
        try:
            rows: AttrDict = self.execute_macro(
                GET_COLUMNS_IN_RELATION_RAW_MACRO_NAME, kwargs={"relation": relation}
            )
            columns = self.parse_describe_extended(relation, rows)
        except dbt.exceptions.DbtRuntimeError as e:
            errmsg = getattr(e, "msg", "")
            found_msgs = (msg in errmsg for msg in TABLE_OR_VIEW_NOT_FOUND_MESSAGES)
            if any(found_msgs):
                pass
            else:
                raise e

        columns = [x for x in columns]
        return columns

    def get_relation(self, database: str, schema: str, identifier: str) -> Optional[BaseRelation]:
        if not self.Relation.get_default_include_policy().database:
            database = None
        return super().get_relation(database, schema, identifier)

    def check_schema_exists(self, database, schema):
        results = self.execute_macro(LIST_SCHEMAS_MACRO_NAME, kwargs={"database": database})

        exists = True if schema in [row[0] for row in results] else False
        return exists

    def list_relations_without_caching(self, schema_relation: ClickZettaRelation) \
            -> List[ClickZettaRelation]:  # type: ignore
        kwargs = {"schema_relation": schema_relation}
        try:
            results = self.execute_macro(LIST_RELATIONS_MACRO_NAME, kwargs=kwargs)
        except DbtDatabaseError as exc:
            if "Object does not exist" in str(exc):
                return []
            raise

        relations = []
        quote_policy = {"database": False, "schema": True, "identifier": True}
        for row in results:
            _schema, _identifier, _is_view, _is_materialized_view = row
            try:
                if _is_view == 'true':
                    _type = RelationType.View
                elif _is_materialized_view == 'true':
                    _type = RelationType.MaterializedView
                else:
                    _type = RelationType.Table
            except ValueError:
                _type = self.Relation.External
            relations.append(
                self.Relation.create(
                    database=None,
                    schema=_schema,
                    identifier=_identifier,
                    quote_policy=quote_policy,
                    type=_type,
                )
            )

        return relations

    def quote_seed_column(self, column: str, quote_config: Optional[bool]) -> str:
        quote_columns: bool = False
        if isinstance(quote_config, bool):
            quote_columns = quote_config
        elif quote_config is None:
            pass
        else:
            msg = (
                f'The seed configuration value of "quote_columns" has an '
                f"invalid type {type(quote_config)}"
            )
            raise CompilationError(msg)

        if quote_columns:
            return self.quote(column)
        else:
            return column

    def timestamp_add_sql(self, add_to: str, number: int = 1, interval: str = "hour") -> str:
        return f"DATEADD({interval}, {number}, {add_to})"

    def valid_incremental_strategies(self):
        return ["append", "merge", "insert_overwrite"]



from dbt.adapters.clickzetta.column import ClickZettaColumn  # noqa
from dbt.adapters.clickzetta.connections import ClickZettaConnectionManager  # noqa
from dbt.adapters.clickzetta.connections import ClickZettaCredentials
from dbt.adapters.clickzetta.relation import ClickZettaRelation  # noqa
from dbt.adapters.clickzetta.impl import ClickZettaAdapter

from dbt.adapters.base import AdapterPlugin  # type: ignore
from dbt.include import clickzetta  # type: ignore

Plugin = AdapterPlugin(
    adapter=ClickZettaAdapter, credentials=ClickZettaCredentials, include_path=clickzetta.PACKAGE_PATH  # type: ignore
)
